from datetime import datetime
from typing import Any, Dict, Optional, List

import requests
from botocore.auth import S3SigV4Auth
from botocore.awsrequest import AWSRequest
from botocore.credentials import Credentials
from prometheus_api_client import PrometheusApiClientException

from prometrix.connect.custom_connect import CustomPrometheusConnect


class AWSPrometheusConnect(CustomPrometheusConnect):
    def __init__(
        self, access_key: str, secret_key: str, region: str, service_name: str, **kwargs
    ):
        super().__init__(**kwargs)
        self._credentials = Credentials(access_key, secret_key)
        self._sigv4auth = S3SigV4Auth(self._credentials, service_name, region)

    def signed_request(
        self, method, url, data=None, params=None, verify=False, headers=None
    ):
        request = AWSRequest(
            method=method, url=url, data=data, params=params, headers=headers
        )
        self._sigv4auth.add_auth(request)
        return requests.request(
            method=method,
            url=url,
            headers=dict(request.headers),
            verify=verify,
            data=data,
        )

    def _custom_query(self, query: str, params: dict = None):
        """
        Send a custom query to a Prometheus Host.

        This method takes as input a string which will be sent as a query to
        the specified Prometheus Host. This query is a PromQL query.

        :param query: (str) This is a PromQL query, a few examples can be found
            at https://prometheus.io/docs/prometheus/latest/querying/examples/
        :param params: (dict) Optional dictionary containing GET parameters to be
            sent along with the API request, such as "time"
        :returns: (list) A list of metric data received in response of the query sent
        :raises:
            (RequestException) Raises an exception in case of a connection error
            (PrometheusApiClientException) Raises in case of non 200 response status code
        """
        params = params or {}
        data = None
        query = str(query)
        # using the query API to get raw data
        response = self.signed_request(
            method="POST",
            url="{0}/api/v1/query".format(self.url),
            data={**{"query": query}, **params},
            params={},
            verify=self.ssl_verification,
            headers=self.headers,
        )
        return response

    def custom_query_range(
        self,
        query: str,
        start_time: datetime,
        end_time: datetime,
        step: str,
        params: Optional[Dict[str, Any]] = None,
    ):
        """
        Send a query_range to a Prometheus Host.
        This method takes as input a string which will be sent as a query to
        the specified Prometheus Host. This query is a PromQL query.
        :param query: (str) This is a PromQL query, a few examples can be found
            at https://prometheus.io/docs/prometheus/latest/querying/examples/
        :param start_time: (datetime) A datetime object that specifies the query range start time.
        :param end_time: (datetime) A datetime object that specifies the query range end time.
        :param step: (str) Query resolution step width in duration format or float number of seconds - i.e 100s, 3d, 2w, 170.3
        :param params: (dict) Optional dictionary containing GET parameters to be
            sent along with the API request, such as "timeout"
        :returns: (dict) A dict of metric data received in response of the query sent
        :raises:
            (RequestException) Raises an exception in case of a connection error
            (PrometheusApiClientException) Raises in case of non 200 response status code
        """
        start = round(start_time.timestamp())
        end = round(end_time.timestamp())
        params = params or {}

        query = str(query)
        response = self.signed_request(
            method="POST",
            url="{0}/api/v1/query_range".format(self.url),
            data={
                **{"query": query, "start": start, "end": end, "step": step},
                **params,
            },
            params={},
            headers=self.headers,
        )
        if response.status_code == 200:
            return response.json()["data"]
        else:
            raise PrometheusApiClientException(
                "HTTP Status Code {} ({!r})".format(
                    response.status_code, response.content
                )
            )

    def get_label_values(self, label_name: str, params: dict = None):
        params = params or {}
        response = self.signed_request(
            method="GET",
            url="{0}/api/v1/label/{1}/values".format(self.url, label_name),
            verify=self.ssl_verification,
            headers=self.headers,
            params=params,
        )
        if response.status_code == 200:
            return response.json()["data"]
        else:
            raise PrometheusApiClientException(
                "HTTP Status Code {} ({!r})".format(
                    response.status_code, response.content
                )
            )

    def all_metrics(self, params: dict = None):
        """
        Get the list of all the metrics that the prometheus host scrapes.

        :param params: (dict) Optional dictionary containing GET parameters to be
            sent along with the API request, such as "time"
        :returns: (list) A list of names of all the metrics available from the
            specified prometheus host
        :raises:
            (RequestException) Raises an exception in case of a connection error
            (PrometheusApiClientException) Raises in case of non 200 response status code
        """
        return self.get_label_values(label_name="__name__", params=params)

    def _send_series(self, data: dict, params: dict) -> requests.Response:
        return self.signed_request(
            method="POST",
            url=f"{self.url}/api/v1/series",
            data=data,
            headers=self.headers,
            params=params,
            verify=self.ssl_verification,
        )

    def get_current_metric_value(self, *args, **kwargs):
        raise NotImplementedError

    def get_metric_range_data(self, *args, **kwargs):
        raise NotImplementedError

    def get_metric_aggregation(self, *args, **kwargs):
        raise NotImplementedError



from typing import Dict, List, Optional

from pydantic import BaseModel

"""
    The Metric object is defined in prometheus as a dictionary so no specific labels are guaranteed
    Known commonly returned labels in dictionary 'metric':
    __name__: the name of the outer function in prometheus
    Other labels 'metric' sometimes contains:
    [container, created_by_kind, created_by_name, endpoint, host_ip, host_network, instance, job, namespace, node, pod,
    pod_ip, service, uid, ...]
"""
PrometheusMetric = Dict[str, str]


class PrometheusScalarValue(BaseModel):
    timestamp: float
    value: str

    def __init__(self, raw_scalar_list: list):
        """
        :var raw_scalar:  is the list prometheus returns as a scalar value from its queries

        While usually this is a list of size 2 in the form of [float_timestamp, str_returned_value]
        the list size and return value types are not guaranteed
        """
        if len(raw_scalar_list) != 2:
            raise Exception(f"Invalid prometheus scalar value {raw_scalar_list}")
        timestamp = float(raw_scalar_list[0])
        value = str(raw_scalar_list[1])
        super().__init__(timestamp=timestamp, value=value)


class PrometheusVector(BaseModel):
    metric: PrometheusMetric
    value: PrometheusScalarValue

    def __init__(self, metric, value):
        super().__init__(metric=metric, value=PrometheusScalarValue(value))

class PrometheusSeries(BaseModel):
    metric: PrometheusMetric
    timestamps: List[float]
    values: List[str]

    def __init__(self, **kwargs):
        raw_values = kwargs.pop("values", None)
        if not raw_values:
            raise Exception("values missing")
        timestamps = []
        values = []
        for raw_value in raw_values:
            prometheus_scalar_value = PrometheusScalarValue(raw_value)
            timestamps.append(prometheus_scalar_value.timestamp)
            values.append(prometheus_scalar_value.value)
        super().__init__(timestamps=timestamps, values=values, **kwargs)


class PrometheusQueryResult(BaseModel):
    """
    This class is the returned object for prometheus queries
    :var result_type:  can be of type "vector", "matrix", "scalar", "string" depending on the query
    :var vector_result:  a formatted vector list from the query result, if the var result_type is "vector"
    :var series_list_result:  a formatted series list from the query result, if the var result_type is "matrix"
    :var scalar_result:  scalar object of the query result, if the var result_type is "scalar"
    :var string_result:  a string of the query result, if the var result_type is "string"

    :raises:
        (Exception) Raises an Exception in the case that there is an issue with the resultType and result not matching
    """

    result_type: str
    vector_result: Optional[List[PrometheusVector]]
    series_list_result: Optional[List[PrometheusSeries]]
    scalar_result: Optional[PrometheusScalarValue]
    string_result: Optional[str]

    def __init__(self, data):
        result = data.get("result", None)
        result_type = data.get("resultType", None)
        vector_result = None
        series_list_result = None
        scalar_result = None
        string_result = None
        if not result_type:
            raise Exception("resultType missing")
        if result is None:
            raise Exception("result object missing")
        elif result_type == "string" or result_type == "error":
            string_result = str(result)
        elif result_type == "scalar" and isinstance(result, list):
            scalar_result = PrometheusScalarValue(result)
        elif result_type == "vector" and isinstance(result, list):
            vector_result = [
                PrometheusVector(**vector_result) for vector_result in result
            ]
        elif result_type == "matrix" and isinstance(result, list):
            series_list_result = [
                PrometheusSeries(**series_list_result) for series_list_result in result
            ]
        else:
            raise Exception("result or returnType is invalid")

        super().__init__(
            result_type=result_type,
            vector_result=vector_result,
            series_list_result=series_list_result,
            scalar_result=scalar_result,
            string_result=string_result,
        )



from prometrix.auth import PrometheusAuthorization
from prometrix.connect.aws_connect import AWSPrometheusConnect
from prometrix.connect.custom_connect import CustomPrometheusConnect
from prometrix.exceptions import (MetricsNotFound,
                                  PrometheusFlagsConnectionError,
                                  PrometheusNotFound, ThanosMetricsNotFound,
                                  VictoriaMetricsNotFound)
from prometrix.models.prometheus_config import (
    AWSPrometheusConfig, AzurePrometheusConfig, CoralogixPrometheusConfig,
    PrometheusApis, PrometheusConfig, VictoriaMetricsPrometheusConfig)
from prometrix.models.prometheus_result import (PrometheusMetric,
                                                PrometheusQueryResult,
                                                PrometheusScalarValue,
                                                PrometheusSeries)
from prometrix.utils import get_custom_prometheus_connect



from datetime import datetime
from typing import Dict, Optional, List

import requests
from prometheus_api_client import (PrometheusApiClientException,
                                   PrometheusConnect)
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError, HTTPError

from prometrix.auth import PrometheusAuthorization
from prometrix.exceptions import (PrometheusFlagsConnectionError,
                                  PrometheusNotFound, VictoriaMetricsNotFound)
from prometrix.models.prometheus_config import PrometheusApis, PrometheusConfig


class CustomPrometheusConnect(PrometheusConnect):
    def __init__(self, config: PrometheusConfig):
        super().__init__(
            url=config.url, disable_ssl=config.disable_ssl, headers=config.headers
        )
        self.config = config
        self._session = requests.Session()
        self._session.mount(self.url, HTTPAdapter(pool_maxsize=10, pool_block=True))

    def custom_query_range(
        self,
        query: str,
        start_time: datetime,
        end_time: datetime,
        step: str,
        params: dict = None,
    ):
        """
        The main difference here is that the method here is POST and the prometheus_cli is GET
        """
        start = round(start_time.timestamp())
        end = round(end_time.timestamp())
        params = params or {}
        data = None
        query = str(query)
        # using the query_range API to get raw data
        response = self._session.post(
            "{0}/api/v1/query_range".format(self.url),
            data={
                "query": query,
                "start": start,
                "end": end,
                "step": step,
                **params,
            },
            verify=self.ssl_verification,
            headers=self.headers,
        )
        if response.status_code == 200:
            return response.json()["data"]
        else:
            raise PrometheusApiClientException(
                "HTTP Status Code {} ({!r})".format(
                    response.status_code, response.content
                )
            )

    def _custom_query(self, query: str, params: dict = None):
        """
        The main difference here is that the method here is POST and the prometheus_cli is GET
        """
        params = params or {}
        data = None
        query = str(query)
        # using the query API to get raw data
        response = self._session.post(
            "{0}/api/v1/query".format(self.url),
            data={"query": query, **params},
            verify=self.ssl_verification,
            headers=self.headers,
        )
        return response

    def get_label_values(self, label_name: str, params: dict = None):
        if PrometheusApis.LABELS not in self.config.supported_apis:
            raise PrometheusApiClientException("Labels Api not supported")
        return super().get_label_values(label_name, params)

    def custom_query(self, query: str, params: dict = None):
        response = self._custom_query(query, params)
        if response.status_code == 200:
            data = response.json()["data"]
        else:
            raise PrometheusApiClientException(
                "HTTP Status Code {} ({!r})".format(
                    response.status_code, response.content
                )
            )
        return data

    def check_prometheus_connection(self, params: dict = None):
        params = params or {}
        try:
            response = self._custom_query(query="example", params=params)
            if response.status_code == 401:
                if PrometheusAuthorization.request_new_token(self.config):
                    self.headers = PrometheusAuthorization.get_authorization_headers(
                        self.config
                    )
                    response = self._custom_query(query="example", params=params)
            response.raise_for_status()
        except (ConnectionError, HTTPError, PrometheusApiClientException) as e:
            raise PrometheusNotFound(
                f"Couldn't connect to Prometheus found under {self.url}\nCaused by {e.__class__.__name__}: {e})"
            ) from e

    def __text_config_to_dict(self, text: str) -> Dict:
        conf = {}
        lines = text.strip().split("\n")
        for line in lines:
            key, val = line.strip().split("=")
            conf[key] = val.strip('"')

        return conf

    def get_prometheus_flags(self) -> Optional[Dict]:
        try:
            if PrometheusApis.FLAGS in self.config.supported_apis:
                return self.fetch_prometheus_flags()
            if PrometheusApis.VM_FLAGS in self.config.supported_apis:
                return self.fetch_victoria_metrics_flags()
        except Exception as e:
            service_name = (
                "Prometheus"
                if PrometheusApis.FLAGS in self.config.supported_apis
                else "Victoria Metrics"
            )
            raise PrometheusFlagsConnectionError(
                f"Couldn't connect to the url: {self.url}\n\t\t{service_name}: {e}"
            )

    def fetch_prometheus_flags(self) -> Dict:
        try:
            response = self._session.get(
                f"{self.url}/api/v1/status/flags",
                verify=self.ssl_verification,
                headers=self.headers,
                # This query should return empty results, but is correct
                params={},
            )
            response.raise_for_status()
            return response.json().get("data", {})
        except Exception as e:
            raise PrometheusNotFound(
                f"Couldn't connect to Prometheus found under {self.url}\nCaused by {e.__class__.__name__}: {e})"
            ) from e

    def fetch_victoria_metrics_flags(self) -> Dict:
        try:
            # connecting to VictoriaMetrics
            response = self._session.get(
                f"{self.url}/flags",
                verify=self.ssl_verification,
                headers=self.headers,
                # This query should return empty results, but is correct
                params={},
            )
            response.raise_for_status()

            configuration = self.__text_config_to_dict(response.text)
            return configuration
        except Exception as e:
            raise VictoriaMetricsNotFound(
                f"Couldn't connect to VictoriaMetrics found under {self.url}\nCaused by {e.__class__.__name__}: {e})"
            ) from e

    def _send_series(self, data: dict, params: dict) -> requests.Response:
        return self._session.post(
            f"{self.url}/api/v1/series",
            data=data,
            verify=self.ssl_verification,
            headers=self.headers,
            params=params,
        )

    def get_series(self, match: List[str], start_time: Optional[datetime] = None,
                   end_time: Optional[datetime] = None, params: dict = None) -> Dict:
        """
        Retrieves a dictionary of series that match the specified label sets from Prometheus.

        :param match: (List[str]) List of string selectors to specify the series to match.
        :param start_time: (Optional[datetime]) The start time for the query as a datetime object.
        :param end_time: (Optional[datetime]) The end time for the query as a datetime object.
        :param params: (Optional[dict]) Additional parameters to be sent in the query.
        :returns: (dict) A dictionary of the query results, which includes the series of matched metrics.
        :raises:
            (PrometheusApiClientException) Raises an exception with details of the response, in case of a non 200 HTTP status code.
        """
        params = params or {}

        # The data to be sent with the POST request
        data = {
            'match[]': match,
        }

        # Include start and end time in the data if provided
        if start_time:
            data['start'] = round(start_time.timestamp())
        if end_time:
            data['end'] = round(end_time.timestamp())

        response = self._send_series(data=data, params=params)
        if response.status_code == 200:
            return response.json()["data"]
        else:
            raise PrometheusApiClientException(
                f"Failed to retrieve `series` data from Prometheus. "
                f"Response status: {response.status_code!r}. "
                f"Response content: {response.content!r}.  "
            )

